
// Particles.java
// Andrew Davison, May 2007, ad@fivedots.coe.psu.ac.th

/* Create and render a particles system.

   Each particle is assigned an initial velocity and a starting
   position in reset(). They are used to calculate its new position 
   based on its age on each call to update(). 

   A particle also has an initial colour (also set in
   reset()). update() gradually increases its transparancy based
   on its age.

   Each particle has its own point size that is used when rendering. The
   point sizes are adjusted on each call to update() based on the
   camera's fractional squared distance from the particle system's 
   source (x,z) position.

   The particles are rendered using point sprites, which are part of
   the GL11 core functionality.
*/

import java.util.*;
import java.nio.*;
import javax.microedition.lcdui.*;
import javax.microedition.khronos.egl.*;
import javax.microedition.khronos.opengles.*;


public class Particles
{
  private static final int NUM_PARTS = 250;    // number of particles

  private static final float GRAVITY = 9.81f; 
                       // for the parabolic vertical distance equation

  private static final float MAX_AGE = 2.5f;
  private static final float AGE_INCR = 0.1f;    
             // particle age increment for each update() call

  // basic x-/z- and y- axis start velocities for a particle
  private static final float VEL_XZ0 = 0.7f;
  private static final float VEL_Y0 = 9.81f;
  private static final float VEL_PERC = 0.2f;  // percentage variation (+/- 20%)

  private static final float MAX_POINT_SIZE = 50.0f;
  private static final float POINT_SIZE_PERC = 0.5f;  // percentage variation (+/- 50%)

  private static final String TEX_FNM = "/part3.png";  // in \res in WTK 2.5


  private GL11 gl;  
    /* use GL11 so OES_point_sprite and OES_point_size_array
       are available */

  private Random rand;

  private float xStart, zStart;       // particles source (x,z) position
  private float[] velX, velY, velZ;   //  velocities in x-, y-, z- directions
  private float[] age;

  private FloatBuffer posnsBuf;    // (x,y,z) positions
  private float[] posns;

  private ByteBuffer colsBuf;  // colors (RGBA)
  private byte[] cols;

  private FloatBuffer sizesBuf;     // point sizes (floats)
  private float[] origSizes, currSizes;
  private float minPointSize, maxPointSize, pointSize;

  private ByteBuffer texBuf;     // RGBA texture image
  private int imWidth, imHeight;   // dimensions of the texture image
  private int texNames[];   // for the texture name

  // camera-related
  private KeyCamera keyCamera;
  private double initialSqrDist;     
          // camera's initial squared distance from the particles source posn
  private float sqrDistFrac;  
         // current distance from source as a fraction of the initial squared dist


  public Particles(GL11 gl, float x, float z, KeyCamera kc)
  {
    this.gl = gl;
    xStart = x;    // store source (x,z) position for the system
    zStart = z;
    keyCamera = kc;

    rand = new Random();

    // set initial squared distance and fraction
    initialSqrDist = keyCamera.sqrDistTo((double)xStart, (double)zStart);
    sqrDistFrac = 1.0f;

    // load the particles texture
    loadTexture(TEX_FNM);

    // generate a texture name
    texNames = new int[1];
    gl.glGenTextures(1, texNames, 0); 

    initParticles();
  }  // end of Particles()



  private void loadTexture(String fnm)
  /* Load the image stored in fnm into a byte buffer.
     This version of loadTexture() stores the image's 
     alpha channel. */
  {
    try {
      Image image = Image.createImage(fnm);
      imWidth = image.getWidth();
      imHeight = image.getHeight();
      int pixels[] = new int[imWidth];   // for a row of pixels

      texBuf = ByteBuffer.allocateDirect(imWidth * imHeight * 4); // RGB _and_ A

      /* Read the image in reverse row order to flip the image
         vertically in the buffer. This will order the texture 
         correctly for OpenGL's coordinate system, where (0,0) 
         is the lower left. */
      for (int i = imHeight-1; i >= 0; i--) {
        image.getRGB(pixels, 0, imWidth, 0, i, imWidth, 1);  // get a pixel row
        for (int j = 0; j < imWidth; j++) {
          // store each pixel in the row as four bytes in RGBA order
          texBuf.put((byte) (pixels[j] >> 16 & 0xff));  // red
          texBuf.put((byte) (pixels[j] >> 8 & 0xff));   // green
          texBuf.put((byte) (pixels[j] >> 0 & 0xff));   // blue
          texBuf.put((byte) (pixels[j] >> 24 & 0xff));  // alpha
        }
      }
      texBuf.rewind();
      System.out.println("Loaded texture from " + fnm);
    }
    catch (Exception e) 
    {  System.out.println("Error loading texture from " + fnm); }
  }  // end of loadTexture()


  private void initParticles()
  {
    velX = new float[NUM_PARTS];
    velY = new float[NUM_PARTS];
    velZ = new float[NUM_PARTS];
    age = new float[NUM_PARTS];

    // create positions array and buffer
    posns = new float[NUM_PARTS * 3];
    ByteBuffer bb = ByteBuffer.allocateDirect(NUM_PARTS * 3*4);  
            // each position is a 3-float (x,y,z) coord
	posnsBuf = bb.asFloatBuffer();

    // create colors array and buffer
    cols = new byte[NUM_PARTS * 4];
    colsBuf = ByteBuffer.allocateDirect(NUM_PARTS * 4);
           // each color is a 4-byte RGBA value

    // create point sizes arrays and buffer
    origSizes = new float[NUM_PARTS];
    currSizes = new float[NUM_PARTS];
    ByteBuffer sb = ByteBuffer.allocateDirect(NUM_PARTS * 4);
    sizesBuf = sb.asFloatBuffer();

    getPointSizesInfo();

    // initialize the particles
    for(int i=0; i< NUM_PARTS; i++)
      reset(i);  

    // fill the particle buffers
	posnsBuf.put(posns).rewind();
    colsBuf.put(cols).rewind();
    sizesBuf.put(currSizes).rewind();
  }  // end of initParticles()


  private void getPointSizesInfo()
  // get the min, max point sizes, and set the default point size
  {
    float[] range = new float[2];
    gl.glGetFloatv(GL11.GL_ALIASED_POINT_SIZE_RANGE, range, 0);
    minPointSize = range[0];
    maxPointSize = (MAX_POINT_SIZE < range[1]) ? MAX_POINT_SIZE : range[1];
               // limit the maximum point size
    // System.out.println("Point size range: " + minPointSize + " - " + maxPointSize);

    // store a default point size
    pointSize = (maxPointSize + minPointSize)/4.0f;
    // System.out.println("Default point size: " + pointSize);
  }  // end of getPointSizesInfo()



  private void reset(int i)
  // assign new values to particle i
  {
    initVels(i); 
    age[i] = 0;

    // set (x,y,z) starting position (the fountain's base)
    posns[i*3] = xStart; 
    posns[i*3+1] = 0;    // always start on the floor
    posns[i*3+2] = zStart;

    // set RGBA colours between 0 and 255
    cols[i*4] = (byte) 255;   // red
    cols[i*4+1] = (byte) 0;  // no green
    cols[i*4+2] = (byte) 0;  // no blue
    cols[i*4+3] = (byte) 200;  // alpha (starts fairly opaque)

    // initialize point size
    origSizes[i] = genPointSize();
    currSizes[i] = scalePointSize( origSizes[i] );
  }  // end of reset()


  private void initVels(int i)
  // initialize velocities in x-, y-, and z- directions
  {
    float velXZ = genVelocity(VEL_XZ0);   // XZ plane velocity

    double rad = Math.toRadians(rand.nextInt(360));  
                        // a random angle between 0-360

    float velX0 = (float)(velXZ * Math.cos(rad));   // vel along x-axis
    float velZ0 = (float)(velXZ * Math.sin(rad));   // vel along z-axis
    // System.out.println("vel(x,z): " + velX0 + ", " + velZ0);
    
    velX[i] = velX0; 
    velY[i] = genVelocity(VEL_Y0);   // vertical velocity
    velZ[i] = velZ0; 
  }  // end of initVels()


  private float genVelocity(float startVel)
  // return startVel +/- VEL_PERC percentage
  {
    float velVar = startVel * rand.nextFloat()*VEL_PERC;
    float vel = (rand.nextInt(2) != 0) ? (startVel + velVar) : (startVel - velVar);
    return vel;
  }  // end of genVelocity()


  private float genPointSize()
  // generate a variation of pointSize +/- POINT_SIZE_PERC percentage
  {
    float psVar = pointSize * rand.nextFloat()*POINT_SIZE_PERC;
    float size = (rand.nextInt(2) != 0) ? (pointSize + psVar) : (pointSize - psVar);
    return size;
  }  // end of genPointSize()


  private float scalePointSize(float sz)
  // scale supplied point size based on squared distance fraction
  {
    float ptSize = sz/sqrDistFrac;
    if (ptSize < minPointSize)
      ptSize = minPointSize;
    else if (ptSize > maxPointSize)
      ptSize = maxPointSize;
    return  ptSize;
  }  // end of scalePointSize()


  public void update()
  // update the values in all the particles
  {
    updateSqrDistFrac();

    // update the particles
    for(int i=0; i< NUM_PARTS; i++) {
      age[i] += AGE_INCR;

      // reset the particle if below the floor (y < 0) or too old
      if ((posns[i*3+1] < 0) || (age[i] > MAX_AGE))
        reset(i); 
      else {
        // vertical position = (v0*t) - (0.5*g*t*t)
        // use age as the time value 
        float vertPos = (velY[i] * age[i]) - (0.5f*GRAVITY * age[i]*age[i]);

        posns[i*3] = xStart + (velX[i] * age[i]);    // x
        posns[i*3+1] = vertPos/2.0f;                 // y with scaling factor 
        posns[i*3+2] = zStart + (velZ[i] * age[i]);  // z

        // reduce alpha based on age (a particle fades away over time)
        cols[i*4+3] = (byte)(255 - (int)(age[i]/MAX_AGE * 255));

        // modify point size
        currSizes[i] = scalePointSize(origSizes[i]);
      }
    }  

    // update the particles buffers
	posnsBuf.put(posns).rewind();
    colsBuf.put(cols).rewind();
    sizesBuf.put(currSizes).rewind();
  }  // end of update()


  private void updateSqrDistFrac()
  // update the camera's squared distance fraction
  {
    double sqrDist = keyCamera.sqrDistTo((double)xStart, (double)zStart);
                                // get current squared distance from camera
    sqrDistFrac = (float) (sqrDist/initialSqrDist);
  } // end of updateSqrDistFrac()



  public void draw()
  // draw the particles as point sprites
  {
    // avoid problem with overlapping alpha blends
    gl.glDepthMask(false);  // protect depth buffer by making it read-only

    gl.glEnable(GL11.GL_BLEND);   // enable colour and texture blending
    gl.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
    gl.glEnable(GL11.GL_TEXTURE_2D);
    gl.glEnable(GL11.GL_COLOR_MATERIAL);

    gl.glEnable(GL11.GL_POINT_SPRITE_OES);

    // generate point sprite texture coords
    gl.glTexEnvx(GL11.GL_POINT_SPRITE_OES, GL11.GL_COORD_REPLACE_OES, GL11.GL_TRUE);
    setTexture();

    // enable buffers for rendering
    gl.glEnableClientState(GL11.GL_VERTEX_ARRAY);
    gl.glEnableClientState(GL11.GL_COLOR_ARRAY);
    gl.glEnableClientState(GL11.GL_POINT_SIZE_ARRAY_OES);

    // load buffers
    gl.glVertexPointer(3, GL11.GL_FLOAT, 0, posnsBuf);
    gl.glColorPointer(4, GL11.GL_UNSIGNED_BYTE, 0, colsBuf);
    gl.glPointSizePointerOES(GL11.GL_FLOAT, 0, sizesBuf);

    gl.glDrawArrays(GL11.GL_POINTS, 0, NUM_PARTS);
  
    // disable the buffers at the end of rendering
    gl.glDisableClientState(GL11.GL_VERTEX_ARRAY);
    gl.glDisableClientState(GL11.GL_COLOR_ARRAY);
    gl.glDisableClientState(GL11.GL_POINT_SIZE_ARRAY_OES);

    // switch off point sprites
    gl.glTexEnvx(GL11.GL_POINT_SPRITE_OES, GL11.GL_COORD_REPLACE_OES, GL11.GL_FALSE);
    gl.glDisable(GL11.GL_POINT_SPRITE_OES);

    // restore other states
    gl.glDisable(GL11.GL_COLOR_MATERIAL);
    gl.glDisable(GL11.GL_TEXTURE_2D);
    gl.glDisable(GL11.GL_BLEND);

    gl.glDepthMask(true);
  }  // end of draw()


  private void setTexture()
  // set the texture used by all the particles at render time
  {
    gl.glBindTexture(GL11.GL_TEXTURE_2D, texNames[0]);  // use the tex name

    // specify the _RGBA_ texture for the currently bound tex name
    gl.glTexImage2D(GL11.GL_TEXTURE_2D, 0, GL11.GL_RGBA, imWidth, imHeight, 0, 
                    GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, texBuf);

    // set the minification/magnification techniques
    gl.glTexParameterx(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
    gl.glTexParameterx(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
  } // end of setTexture()

}  // end of Particles class
