
// ParticlesGameCanvas.java
// Andrew Davison, May 2007, ad@fivedots.coe.psu.ac.th

/* The canvas displays a 3D scene overlaid with 2D textual info.
   The 3D scene (created with OpenGL ES) is a floor containing
   a fountain of particles, with a textured cube in the background.

   The camera can be moved around the scene via keyboard controls.
*/


import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;

import javax.microedition.khronos.egl.*;
import javax.microedition.khronos.opengles.*;



public class ParticlesGameCanvas extends GameCanvas implements Runnable
{
  private static final int PERIOD = 250;    // frame rate (in ms)
  private static final float[] LIGHT_DIR = { 1.0f, 1.0f, 1.0f, 0.0f };   
          // right, top, front directional light

  // graphics related
  private Graphics g2d;  // 2D graphics
  private EGL11 egl;     // OpenGL ES link from Java
  private GL11 gl;       // for calling JOGL-ES
  private EGLDisplay eglDisplay; 
  private EGLContext eglContext;     // OpenGL ES state
  private EGLSurface eglWinSurface;  // for on-screen rendering 

  // rendering related
  private int width, height;  // dimensions of the device's screen
  private volatile boolean isRunning;
  private long frameDuration;  // in ms
      // how long one iteration of the animation loop takes

  // scene related
  private Font msgFont;
  private Floor floor;
  private KeyCamera keyCamera;

  private Particles particles; 
  private TexCube texCube;


  public ParticlesGameCanvas()
  {
    super(true);

    width = getWidth();
    height = getHeight();

    msgFont = Font.getFont(Font.FACE_PROPORTIONAL, 
                          Font.STYLE_BOLD, Font.SIZE_LARGE);

    // start the rendering thread
    Thread thread = new Thread(this);
    isRunning = true;
    thread.start();
  } // end of ParticlesGameCanvas()


  public void run()
  /* Contains three stages: initialization, animation loop, shutdown.
     The animation loop has three parts: update, draw, maybe sleep.
  */
  {
    if (!initGraphics())
      return;   // give up if there's an error during initialization

    initScene();

    long startTime;
    while (isRunning) {
      startTime = System.currentTimeMillis();

      keyCamera.update( getKeyStates() );
      particles.update();

      drawScene();

      frameDuration = System.currentTimeMillis() - startTime;
      try { // sleep a bit maybe, so one iteration takes PERIOD ms
        if (frameDuration < PERIOD)
          Thread.sleep(PERIOD - (int)frameDuration); 
      }
      catch (InterruptedException e){}
    }

    shutdown();
  } // end of run()


  public void stopRunning()
  // called from the MIDlet at termination time
  { isRunning = false; }


  private boolean initGraphics()
  /* Set up the 2D and 3D graphics for this thread.
     This version of initGraphics() has much more error checking. */
  {
    // initialize OpenGL ES
    egl = (EGL11) EGLContext.getEGL();
    if (egl == null) {
      System.out.println("Error: could not initialize OpenGL ES");
      return false;
    }

    // initialize the OpenGL ES connection to the display
    eglDisplay = egl.eglGetDisplay(EGL11.EGL_DEFAULT_DISPLAY);
    if (eglDisplay == null) {
      System.out.println("Error: could not open connection to display");
      return false;
    }

    int[] majorMinor = new int[2];
    if (!egl.eglInitialize(eglDisplay, majorMinor)) {
      System.out.println("Error: could not initialize OpenGL ES Display");
      return false;
    }
    System.out.println("EGL version: " + majorMinor[0] + "." + majorMinor[1]);

    // determine the number of available configurations
    int[] numConfigs = new int[1];
    egl.eglGetConfigs(eglDisplay, null, 0, numConfigs);
    if (numConfigs[0] < 1) {
      System.out.println("Error: no configurations found");
      return false;
    }

    // specify an 8/8/8/8 RGBA configuration
    int configAttributes[] = {
      EGL11.EGL_RED_SIZE, 8, EGL11.EGL_GREEN_SIZE, 8, EGL11.EGL_BLUE_SIZE, 8,  // RGB
      EGL11.EGL_ALPHA_SIZE, 8, // EGL11.EGL_DONT_CARE, // 8, // alpha
      EGL11.EGL_DEPTH_SIZE, 16,                     // use a 16-bit z-buffer
      // EGL11.EGL_STENCIL_SIZE, EGL11.EGL_DONT_CARE,   // don't care about stencils
      EGL11.EGL_SURFACE_TYPE, EGL11.EGL_WINDOW_BIT, // use a window buffer
      EGL11.EGL_NONE
    };

    // use the first matching configuration
    EGLConfig eglConfigs[] = new EGLConfig[numConfigs[0]];
    if (!egl.eglChooseConfig(eglDisplay, configAttributes, eglConfigs,
                        eglConfigs.length, numConfigs)) {
      System.out.println("Error: could not find a suitable configuration");
      return false;
    }
    EGLConfig eglConfig = eglConfigs[0];

    /* initialize the OpenGL ES rendering state (the context) 
       with the display and configuration */
    eglContext =  egl.eglCreateContext(eglDisplay, eglConfig, 
                                                 EGL11.EGL_NO_CONTEXT, null);
    if (eglContext == null) {
      System.out.println("Error: could not create a rendering state");
      return false;
    }

    // initialize 2D and 3D graphics: the APIs are called through these objects
    g2d = getGraphics();   // 2D
    gl = (GL11) eglContext.getGL();  // 3D
    if (gl == null) {
      System.out.println("Error: could not create a 3D graphics context");
      return false;
    }

    // set the drawing surface to be a window (i.e. use on-screen rendering)
    eglWinSurface = egl.eglCreateWindowSurface(eglDisplay, eglConfig, g2d, null);
    if (eglWinSurface == null) {
      System.out.println("Error: could not create a drawing surface window");
      return false;
    }

    // bind the display, drawing surface, and context to this thread  
    if (!egl.eglMakeCurrent(eglDisplay, eglWinSurface, eglWinSurface, eglContext)){
      System.out.println("Error: could not make the context current");
      return false;
    }

    return true;   // everything worked!
  } // end of initGraphics()



  private void initScene()
  // initialize the view, camera, lighting, and create the scenery
  {
    setView(60.0f, 0.1f, 50.0f); 

    keyCamera = new KeyCamera(gl);

    gl.glClearColor(0.17f, 0.65f, 0.92f, 1.0f);  // sky blue background

    // z- (depth) selectBuffer initialization for hidden surface removal
    gl.glEnable(GL11.GL_DEPTH_TEST); 

    gl.glHint(GL11.GL_PERSPECTIVE_CORRECTION_HINT, GL11.GL_NICEST);
    gl.glShadeModel(GL11.GL_SMOOTH);    // use smooth shading

    addLight();
    createScenery();
  }  // end of initScene()


  private void setView(float fovy, float near, float far)
  /* Set the perspective view: the arguments are the field-of-view,
     and the near & far clipping planes */
  {
    gl.glViewport(0, 0, width, height);    
           // set size of drawing area to be the screen size

    // set the viewing frustum (similar to gluPerspective())
    float aspectRatio = (float)width/(float)height;

    float top = near * (float)Math.tan((fovy*Math.PI)/360.0);
    float bottom = -top;
    float left = bottom * aspectRatio;
    float right = top * aspectRatio;

    gl.glMatrixMode(GL11.GL_PROJECTION);  //  set the view attributes
    gl.glLoadIdentity();
    gl.glFrustumf(left, right, bottom, top, near, far); 
  } // end of setView()


  private void addLight()
  // set the lighting properties
  {
    gl.glEnable(GL11.GL_LIGHTING);
    gl.glEnable(GL11.GL_LIGHT0);
    gl.glEnable(GL11.GL_NORMALIZE);

    float[] ambientLight = { 1.0f, 1.0f, 1.0f, 1.0f };     // white ambient
    gl.glLightfv(GL11.GL_LIGHT0, GL11.GL_AMBIENT, ambientLight, 0);

    float[] diffuseLight = { 0.9f, 0.9f, 0.9f, 1.0f };   // white diffuse
    gl.glLightfv(GL11.GL_LIGHT0, GL11.GL_DIFFUSE, diffuseLight, 0);
  } // end of addLight()


  private void createScenery()
  // create the floor, the particles, and the textured cube
  {
    // gl.glEnable(GL11.GL_CULL_FACE);   
              // to cull back faces (a useful optimization)

    floor = new Floor(gl, "/bigGrid.png", 8);   // 8 by 8 size

    particles = new Particles(gl, 0.0f, 0.0f, keyCamera);  
                // positioned at the origin on the floor

    texCube = new TexCube(gl, "/brick2.png", 0, 0.51f, -2.0f, 0.5f);  // (x,y,z), scale
  } // end of createScenery()



  private void drawScene()
  // render the 3D graphics then switch to 2D for the 'overlay' info
  {
    drawSceneGL();    // draw using JOGL-ES

    // 2D graphics
    // show the camera's position, rotations, and mode
    g2d.setColor(0xff0000);   // red
    g2d.setFont(msgFont);

    g2d.drawString( keyCamera.getPos(), 5, 5, Graphics.TOP|Graphics.LEFT);
    g2d.drawString( keyCamera.getRots(), 5, 23, Graphics.TOP|Graphics.LEFT);
    g2d.drawString( keyCamera.getMode(), 5, 41, Graphics.TOP|Graphics.LEFT);

    // on the right, show the current frame duration
    g2d.setColor(0x000000);   // black
    g2d.drawString( "" + frameDuration, width - 30, 5, Graphics.TOP|Graphics.LEFT);

    flushGraphics();
  }  // end of drawScene()


  private void drawSceneGL()
  // draw the scene (the camera, the light source, floor, the models)
  {
    // wait until OpenGL ES is available before starting to draw
    egl.eglWaitNative(EGL11.EGL_CORE_NATIVE_ENGINE, g2d);

    // clear colour and depth buffers
    gl.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);

    // set modeling and viewing transformations
    gl.glMatrixMode(GL11.GL_MODELVIEW);
    gl.glLoadIdentity();

	keyCamera.position();   // position the camera

    // set light direction
    gl.glLightfv(GL11.GL_LIGHT0, GL11.GL_POSITION, LIGHT_DIR, 0);

    floor.draw();
    texCube.draw();

    particles.draw();   // draw alpha blended stuff last

    // wait until all Open GL ES tasks are finished
    gl.glFinish();  
    egl.eglWaitGL();
  } // end of drawSceneGL()



  private void shutdown()
  // close down OpenGL ES
  {
    if ((egl == null) || (eglDisplay == null))
      return;

    // disconnect the display, drawing surface, and context from this thread  
    egl.eglMakeCurrent(eglDisplay, EGL11.EGL_NO_SURFACE, 
                         EGL11.EGL_NO_SURFACE, EGL11.EGL_NO_CONTEXT);

    // delete the context
    if (eglContext != null)
      egl.eglDestroyContext(eglDisplay, eglContext);

    // delete the drawing surface
    if (eglWinSurface != null)
      egl.eglDestroySurface(eglDisplay, eglWinSurface);

    // break the OpenGL ES connection to the display
    egl.eglTerminate(eglDisplay);
  }  // end of shutdown()

} // end of ParticlesGameCanvas class
